/*-
 * Copyright (C)2007 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2007 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: pgsadump.c,v 1.8 2008/05/04 12:55:54 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include "pgsadump.h"

static void
usage(PGsaDumpCtl *ctl, int ex) {
    int i;

    fprintf( stderr,
             "pgsadump " VERS "  " DATE " @BABOLO http://www.babolo.ru/\n"
             "pgsadump[ -flags] file[ [_]type..]\n"
             " -0 NULL bitmask output\n"
             " -1 stored ctid output\n"
             " -2 oid output\n"
             " -3 xmin output\n"
             " -4 cmin output\n"
             " -5 xmax output\n"
             " -6 cmax output\n"
             " -7 t_infomask output\n"
             " -8 natts output\n"
             " -9 real ctid output\n"
             " -a min natts to output\n"
             " -A max natts to output\n"
             " -b block size\n"
             " -B do not stop on block structure errors\n"
             " -c do not stop on attribute structure errors\n"
             " -e tuples with stored ctid <> real ctid output\n"
             " -h help\n"
             " -l tuples without LP_USED output\n"
             " -L tuples with LP_DELETE output\n"
             " -m term to output\n"
             " -M mask to output\n"
             " -n layout version\n"
             " -N NBASE for numeric\n"
             " -p multiplicator for -P\n"
             " -P ctid shift\n"
             " -q quiet\n"
             " -S do not stop on block size mismatch\n"
             " -t do not stop on tuple structure errors\n"
             " -vvv verbose\n"
             " -V do not stop on layout version mismatch\n"
             " -x NULL output when natts less then intended\n"
             " -X MAXIMUM_ALIGNOF\n"
             "types are:\n"
           );
    for (i = 0; PGsaDumpCollection[ctl->pd_ver][i].conv; i++)
        fprintf(stderr, "  %s\n", PGsaDumpCollection[ctl->pd_ver][i].descr);
    exit(ex);
}

int
main(int argc, char **argv, char **envp) {
    mife_descriptor *in = NULL;
    int ex = EX_OK, ch, i, j;
    u_int32_t pgn, lfag = 0, pgmul = 131072, pgplus = 0;
    off_t pos;
    ssize_t last = 0;
    PGsaDumpCtl ctl;

    ctl.flags = BLIN_VER1 | PGsaDumpAVER | PGsaDumpABSZ;
    ctl.blksz = 8192;
    ctl.nbase = 10000;
    ctl.ind_maxk = 32;
    ctl.atrmax = 1664;
    ctl.atrmin = 0;
    ctl.atrterm = 0;
    ctl.atrmask = 0;
    ctl.pd_ver = 1;
    ctl.alignmax = sizeof(int);
    while ((ch = getopt(argc, argv, "0123456789a:A:b:BceEhi:lLm:M:n:N:p:P:qStvVxX:")) != EOF)
        switch (ch) {
         case '0': ctl.flags |= PGsaDumpNULS;               break;
         case '1': ctl.flags |= PGsaDumpCTID;               break;
         case '2': ctl.flags |= PGsaDumpOID;                break;
         case '3': ctl.flags |= PGsaDumpXMIN;               break;
         case '4': ctl.flags |= PGsaDumpCMIN;               break;
         case '5': ctl.flags |= PGsaDumpXMAX;               break;
         case '6': ctl.flags |= PGsaDumpCMAX;               break;
         case '7': ctl.flags |= PGsaDumpINFO;               break;
         case '8': ctl.flags |= PGsaDumpNATR;               break;
         case '9': ctl.flags |= PGsaDumpRTID;               break;
         case 'a': ctl.atrmin = strtoul(optarg, NULL, 0);   break;
         case 'A': ctl.atrmax = strtoul(optarg, NULL, 0);   break;
         case 'b': ctl.flags &= ~PGsaDumpABSZ;
                   ctl.blksz = strtoul(optarg, NULL, 0);    break;
         case 'B': lfag = 1;                                break;
         case 'c': ctl.flags |= PGsaDumpXCOL;               break;
         case 'e': ctl.flags |= PGsaDumpETID;               break;
         case 'E': ctl.flags |= PGsaDumpHEXT;               break;
         case 'h': usage(&ctl, ex);                         break;
         case 'i': ctl.ind_maxk = strtoul(optarg, NULL, 0); break;
         case 'l': ctl.flags |= PGsaDumpLNUS;               break;
         case 'L': ctl.flags |= PGsaDumpLDEL;               break;
         case 'm': ctl.atrterm = strtoul(optarg, NULL, 0);  break;
         case 'M': ctl.atrmask = strtoul(optarg, NULL, 0);  break;
         case 'n': ctl.flags &= ~PGsaDumpAVER;
                   ctl.pd_ver = strtoul(optarg, NULL, 0);   break;
         case 'N': ctl.nbase = strtoul(optarg, NULL, 0);    break;
         case 'p': pgmul = strtoul(optarg, NULL, 0);        break;
         case 'P': pgplus = pgmul * strtoul(optarg, NULL, 0);break;
         case 'q': setBLIN_VERMASK(ctl.flags, 0);           break;
         case 'S': ctl.flags |= PGsaDumpXBSZ;               break;
         case 't': ctl.flags |= PGsaDumpXTUP;               break;
         case 'v': BLIN_VERBOSE(ctl.flags);                 break;
         case 'V': ctl.flags |= PGsaDumpXVER;               break;
         case 'x': ctl.flags |= PGsaDumpXATR;               break;
         case 'X': ctl.alignmax = strtoul(optarg, NULL, 0); break;
         default:  usage(&ctl, EX_USAGE);
        }
    argc -= optind;
    argv += optind;
    if  (argc > 0 && !argv[0]) usage(&ctl, EX_USAGE);
    if  (!(in = mife_open(kukMIFE_BUFL(1), argv[0]))) {
        ifBLIN_QV1(ctl.flags) warn("file %s", argv[0]);
        exit(EX_NOINPUT);
    }
    ctl.atrs[0] = 0;
    for (i = 1; !ex && i < argc && argv[i]; i++) {
        ex = EX_DATAERR;
        if  (argv[i][0] == '_') {
            for (j = 0; PGsaDumpCollection[ctl.pd_ver][j].conv; j++) {
                if  (!strcmp(&(argv[i][1]), PGsaDumpCollection[ctl.pd_ver][j].descr)) {
                    ctl.atrs[i] = j | 0x8000;
                    ex = EX_OK;
                    break;
            }   }
        } else {
            for (j = 0; PGsaDumpCollection[ctl.pd_ver][j].conv; j++) {
                if  (!strcmp(argv[i], PGsaDumpCollection[ctl.pd_ver][j].descr)) {
                    ctl.atrs[i] = j;
                    ex = EX_OK;
                    break;
        }   }   }
        if  (ex) {ifBLIN_QV1(ctl.flags) fprintf(stderr, "Illegal type: %s\n", argv[i]);}
          else ctl.atrs[0] = i;
    }
    ifBLIN_QV3(ctl.flags) {
        for (i = 0; i <= ctl.atrs[0]; i++) {
            fprintf(stderr, "  [%d]=%d\n", i, ctl.atrs[i]);
    }   }
    for (pos = 0, pgn = 0;; pos += ctl.blksz) {
        if  ((last = mife_read(in, ctl.blksz, pos)) < 0) {
            ifBLIN_QV1(ctl.flags) warn("input");
            ex = EX_DATAERR;
            break;
        }
        if  (last < ctl.blksz) break;
        ifBLIN_QV3(ctl.flags) fprintf(stderr, "pageN=%d, offset=%lld\n", pgn, pos);
        ex = PGsaDumpBlock(&ctl, pgplus + pgn++, mife_get(in, pos));
        if  (!lfag && ex) break;
    }
    if  (last != 0) {
        ifBLIN_QV2(ctl.flags) warnx("truncated");
        ex = EX_DATAERR;
    }
    mife_close(in);
    exit(ex);
}
